/*
 *  Copyright (C) 2002-2006  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <ctype.h>

#include "dosbox.h"

#include "support.h"
#include "inout.h"
#include "pic.h"
#include "setup.h"
#include "timer.h"
#include "bios.h"					// SetLPTPort(..)
#include "hardware.h"				// OpenCaptureFile

#include "parport.h"
#include "directlpt_win32.h"
#include "directlpt_linux.h"
#include "printer_redir.h"
#include "filelpt.h"
#include "dos_inc.h"
#include "tcplpt.h"
#include "printer_direct.h"
#include "sysmenu.h"
#include "printer_charmaps.h"

uint32_t const parallelPortsMax = 3;
device_LPT* parallelPorts[parallelPortsMax] = { 0 };

Bit8u const device_LPT::defaultIRQ[] = { 7, 5, 12 };

static void Parallel_EventHandler(Bitu val) {
	Bitu serclassid=val&0x3;
	if (parallelPorts[serclassid])
		parallelPorts[serclassid]->DeviceGetActive()->handleEvent(val>>2);
}

bool device_LPT::Read(Bit8u* const data, Bit16u* const size) {
	*size=0;
	LOG(LOG_DOSMISC,LOG_NORMAL)("LPTDEVICE:Read called");
	return true;
}

bool device_LPT::Write(Bit8u* const data, Bit16u* const size)
{
    auto const device = DeviceGetActive();
	for (Bit16u i = 0; i < *size; i++) {
		if(!device->Putchar(data[i])) return false;
	}
    
	return true;
}

bool device_LPT::Seek(Bit32u* const pos, Bit32u /*type*/) { *pos = 0; return true; }
bool device_LPT::Close()
{
    auto const device = DeviceGetActive();
    device->handleClose();
    return false;
}

Bit16u device_LPT::GetInformation()
{ return 0x80A0 | (DeviceGetActive()->bReportIOStatusAsEOF ? 0x40 : 0); };


device_LPT::device_LPT(Bit8u const portNum, CParallel* const device) : _num(portNum)
                                                                     , _pportclass(device)
{
    char const* const lptname[]={"LPT1","LPT2","LPT3"};
    static_assert(LENGTH_OF(lptname) >= parallelPortsMax, "Insufficent port names.");
    
    assert(portNum <= parallelPortsMax);
    SetName(lptname[portNum]);
    
    DOS_AddDevice(this);
}

device_LPT::~device_LPT() {
    // Cannot call DOS_DelDevice b/c they delete us.
    // Usually the DOS_Device obj doesn't invoke DOS_DelDevice, but b/c of the dual-dispatch
    // nature from the redirect feature, we have to handle this ourselves, instead of by our
    // port class impl(s).
    for (Bitu i = 0; i < DOS_DEVICES; ++i) {
		if (!Devices[i]                        ) continue;
        if ( strcasecmp(Devices[i]->name, name)) continue;
        Devices[i] = nullptr;
        break;
    }
}

CParallel* device_LPT::DeviceGetBase    () { return _pportclass    .get(); }
CParallel* device_LPT::DeviceGetRedirect() { return _redirectDevice.get(); }
CParallel* device_LPT::DeviceGetActive  () { return DeviceGetRedirect() ? DeviceGetRedirect()
                                                                        : DeviceGetBase(); }

void device_LPT::RedirectEnable(bool const enabled, uint32_t port)
{
    if ((!!_redirectDevice) == enabled) return;
    /// \todo Instead of flushing events, add a seperate queue/entry/id.
    ///       Beware of side-effects when/if this is added.
    auto const eventMask = 0x3u;
    assert(_num <= eventMask); //< Ensure we correctly mask the port #
    PIC_RemoveSpecificEventsMasked(Parallel_EventHandler, eventMask, _num);
    
    if (!enabled) {
        _redirectDevice.reset();
        return;
    }
    
    char paramPort[32]; sprintf_s(paramPort, "port:%u", port);
    char paramCP  [32]; sprintf_s(paramCP  , "cp:%u"  , _pportclass->_codepage);
    char const* const params[] = {"INTERNAL", "ip:127.0.0.1", paramPort
                                 ,"timeout:1000", paramCP, "addLF", "reporteof"};
    CommandLine settings(LENGTH_OF(params), params);
    
    _redirectDevice.reset(new CTCPLPT(&settings, _num, _pportclass->irq));    
}


void CParallel::setEvent(Bit16u type, float duration) {
    assert(port_nr <= 0x3); //< Overflow of port # into event section.
    PIC_AddEvent(Parallel_EventHandler,duration,(type<<2)|port_nr);
}

void CParallel::removeEvent(Bit16u type) {
    assert(port_nr <= 0x3); //< Overflow of port # into event section.
    // TODO
	PIC_RemoveSpecificEvents(Parallel_EventHandler,(type<<2)|port_nr);
}

void CParallel::handleEvent(Bit16u type) {
	handleUpperEvent(type);
}

static Bitu PARALLEL_Read (Bitu port, Bitu iolen) {
	for(Bitu i = 0; i < 3; i++) {
		if(parallel_baseaddr[i]==(port&0xfffc) && parallelPorts[i]) {
            auto const obj = parallelPorts[i]->DeviceGetActive();
			Bitu retval=0xff;
			switch (port & 0x7) {
				case 0: retval = obj->Read_PR (); break;
				case 1: retval = obj->Read_SR (); break;
				case 2: retval = obj->Read_COM(); break;
			}

#if PARALLEL_DEBUG
			const char* const dbgtext[]= {"DAT","STA","COM","???"};
			obj->log_par(obj->dbg_cregs, "read  0x%2x from %s.", retval, dbgtext[port&3]);
#endif
			return retval;	
		}
	}
	return 0xff;
}

static void PARALLEL_Write (Bitu port, Bitu val, Bitu) {
	for(Bitu i = 0; i < 4; i++) {
		if(parallel_baseaddr[i]==(port&0xfffc) && parallelPorts[i]) {
            auto const obj = parallelPorts[i]->DeviceGetActive();
#if PARALLEL_DEBUG
			const char* const dbgtext[]={"DAT","IOS","CON","???"};
			obj->log_par(obj->dbg_cregs,
				"write 0x%2x to %s.",val,dbgtext[port&3]);
			if (obj->dbg_plaindr &&!(port & 0x3)) {
				fprintf(obj->debugfp,"%c",val);
			}
#endif
			switch (port & 0x3) {
				case 0: obj->Write_PR   (val); return;
				case 1: obj->Write_IOSEL(val); return;
				case 2: obj->Write_CON  (val); return;
			}
		}
	}
}

//The Functions

#if PARALLEL_DEBUG
#include <stdarg.h>
void CParallel::log_par(bool active, char const* format,...) {
	if(active) {
		// copied from DEBUG_SHOWMSG
		char buf[512];
		buf[0]=0;
		sprintf(buf,"%12.3f ",PIC_FullIndex());
		va_list msg;
		va_start(msg,format);
		vsprintf(buf+strlen(buf),format,msg);
		va_end(msg);
		// Add newline if not present
		Bitu len=strlen(buf);
		if(buf[len-1]!='\n') strcat(buf,"\r\n");
		fputs(buf,debugfp);
        fflush(debugfp);
	}
}
#endif

// Initialisation
CParallel::CParallel(CommandLine* cmd, Bitu portnr, Bit8u initirq) {
    this->bReportIOStatusAsEOF = cmd->FindExist("reporteof", false);
    
	base = parallel_baseaddr[portnr];
	irq = initirq;
	port_nr = portnr;
    _codepage = 0;
    
    std::string tmp;
    if (cmd->FindStringBegin("cp:", tmp, false))
        sscanf(tmp.c_str(), "%u", &_codepage);
    
#if PARALLEL_DEBUG
	dbg_data	= cmd->FindExist("dbgdata", false);
	dbg_putchar = cmd->FindExist("dbgput", false);
	dbg_cregs	= cmd->FindExist("dbgregs", false);
	dbg_plainputchar = cmd->FindExist("dbgputplain", false);
	dbg_plaindr = cmd->FindExist("dbgdataplain", false);
	
	if(cmd->FindExist("dbgall", false)) {
		dbg_data= 
		dbg_putchar=
		dbg_cregs=true;
		dbg_plainputchar=dbg_plaindr=false;
	}

	if(dbg_data||dbg_putchar||dbg_cregs||dbg_plainputchar||dbg_plaindr)
		debugfp=OpenCaptureFile("parlog",".parlog.txt");
	else debugfp=0;

	if(debugfp == 0) {
		dbg_data= 
		dbg_putchar=dbg_plainputchar=
		dbg_cregs=false;
	} else {
		std::string cleft;
		cmd->GetStringRemain(cleft);

		log_par(true,"Parallel%d: BASE %xh, initstring \"%s\"\r\n\r\n",
			portnr+1,base,cleft.c_str());
	}
#endif
	LOG_MSG("Parallel%d: BASE %xh",portnr+1,base);

	for (Bitu i = 0; i < 3; i++) {
		WriteHandler[i].Install (i + base, PARALLEL_Write, IO_MB);
		ReadHandler[i].Install (i + base, PARALLEL_Read, IO_MB);
	}
	BIOS_SetLPTPort(portnr,base);
	
};

CParallel::~CParallel(void) {
	BIOS_SetLPTPort(port_nr, 0);
};

Bit8u CParallel::getPrinterStatus()
{
	/*	7      not busy
		6      acknowledge
		5      out of paper
		4      selected
		3      I/O error
		2-1    unused
		0      timeout  */
	Bit8u statusreg=Read_SR();
#if PARALLEL_DEBUG
	LOG_MSG("get printer status: %x",statusreg);
#endif
    
	statusreg^=0x48;
    statusreg&=~0x7;
	return statusreg;
}

#include "callback.h"

void RunIdleTime(Bitu milliseconds)
{
	Bitu time=SDL_GetTicks()+milliseconds;
	while(SDL_GetTicks()<time)
		CALLBACK_Idle();
}

void CParallel::initialize()
{
	Write_IOSEL(0x55); // output mode
	Write_CON(0x08); // init low
	Write_PR(0);
	RunIdleTime(10);
	Write_CON(0x0c); // init high
	RunIdleTime(500);
	//LOG_MSG("printer init");
}


class PARPORTS:public Module_base {
public:
	
	PARPORTS (Section * configuration):Module_base (configuration) {

#if C_PRINTER
		bool printer_used = false;
#endif

		Section_prop *section = static_cast <Section_prop*>(configuration);
		
		char pname[]="parallelx";
		// iterate through all 3 lpt ports
		for (Bitu i = 0; i < parallelPortsMax; i++) {
            parallelPorts[i] = nullptr;
            
            static_assert(parallelPortsMax <= 9, "Change port name gen code to handle.");
			pname[8] = '1' + i;
			CommandLine cmd(0,section->Get_string(pname));
            
			std::string str; cmd.FindCommand(1,str);
            if (str == "disabled") continue;
            if (str == "file"    ) { parallelPorts[i] = device_LPT::Create<CFileLPT  >(i, &cmd); continue; }
			if (str == "tcp"     ) { parallelPorts[i] = device_LPT::Create<CTCPLPT   >(i, &cmd); continue; }
#ifdef C_DIRECTLPT			
            if (str == "reallpt" ) { parallelPorts[i] = device_LPT::Create<CDirectLPT>(i, &cmd); continue; }
#endif
            if (str == "direct"  ) { parallelPorts[i] = device_LPT::Create<PrinterDirect>(i, &cmd); continue; }
#if C_PRINTER
			if (str == "printer") {
				if (printer_used) {
					LOG_MSG("Error: Only one port may use \"printer\".");
                    continue;
				}
                
                parallelPorts[i] = device_LPT::Create<CPrinterRedir>(i, &cmd);
                printer_used = !!parallelPorts[i];
				if (!parallelPorts[i])
                    LOG_MSG("Error: printer is not enabled.");
                
                continue;
            }
#endif				
			
            
			LOG_MSG ("Invalid type for LPT%d.", i + 1);
        }
	}

    ~PARPORTS()
    {
        for (auto& port : parallelPorts) {
            delete port;
            port = nullptr;
        }
    }
};

static PARPORTS *testParallelPortsBaseclass;

void PARALLEL_Destroy (Section * sec) {
	delete testParallelPortsBaseclass;
	testParallelPortsBaseclass = NULL;
}

void PARALLEL_Init (Section * sec) {
	// should never happen
	if (testParallelPortsBaseclass) delete testParallelPortsBaseclass;
	testParallelPortsBaseclass = new PARPORTS (sec);
	sec->AddDestroyFunction (&PARALLEL_Destroy, true);
}
