/*
 *  Copyright (C) 2002-2006  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "dosbox.h"

#include "parport.h"
#include "tcplpt.h"
#include "callback.h"
#include "pic.h"
#include <stdio.h>

#include "printer_charmaps.h"

CTCPLPT::CTCPLPT(CommandLine* cmd, Bitu nr, Bit8u initIrq)
                : CParallel(cmd, nr, initIrq)
                , port(0)
                , addLF(false)
                , lastChar(0)
	            , codepage_ptr(nullptr)
                , datareg(0)
	            , controlreg(0)
                , autofeed(false)
	            , ack(false)
	            , timeout(500)
                , lastUsedTick(0)
{
    InstallationSuccessful = false;
    
    std::string tmp;
    
    if (!cmd->FindStringBegin("ip:", ip, false)) {
        LOG_MSG("parallel%d: No IP specified.",nr+1); return;
    }
    
    std::string portname;
    if (!cmd->FindStringBegin("port:",portname,false)) {
        LOG_MSG("parallel%d: No port specified.", nr + 1); return;
    }
    if (sscanf(portname.c_str(), "%u", &port) != 1) {
        LOG_MSG("parallel%d: Invalid port parameter.", nr + 1); return;
    }
    
    if (cmd->FindStringBegin("timeout:", tmp, false)) {
		if (sscanf(tmp.c_str(), "%u", &timeout) != 1) {
			LOG_MSG("parallel%d: Invalid timeout parameter.", nr + 1); return;
		}
	}
    
	addLF = cmd->FindStringBegin("addLF", tmp, false);
	
	// find the codepage
	if (cmd->FindStringBegin("cp:", tmp, false)) {
        Bitu temp = 0;
		if (sscanf(tmp.c_str(), "%u", &temp) != 1) {
			LOG_MSG("parallel%d: Invalid codepage parameter.", nr + 1); return;
		}
        
		for (auto i = 0; charmap[i].codepage; ++i) {
			if (charmap[i].codepage != temp) continue;
            
			codepage_ptr = charmap[i].map;
			break;
        }
	}
    
	InstallationSuccessful = true;
}

CTCPLPT::~CTCPLPT () { removeEvent(0); } //< remove tick handler


bool CTCPLPT::Putchar(Bit8u val)
{	
#if PARALLEL_DEBUG
	log_par(dbg_putchar,"putchar  0x%2x",val);
	if(dbg_plainputchar) fprintf(debugfp,"%c",val);
#endif
    
	if (!OpenSocket()) return false;
	// write to file (or not)
	lastUsedTick = PIC_Ticks;
	
	if (codepage_ptr) {
		auto const extchar = codepage_ptr[val];
		if (extchar & 0xFF00)
            _socket->Write(char(extchar >> 8));
        
        _socket->Write(char(extchar & 0xFF));
	} else {
        _socket->Write(char(val));
    }
    
	if (addLF) {
		if ((lastChar == 0x0d) &&
            (val      != 0x0a)) 
           _socket->Write(char(0x0a));
        
		lastChar = val;
	}
    
	return true;
}

Bitu CTCPLPT::Read_PR () { return datareg; }
Bitu CTCPLPT::Read_COM() { return       0; }
Bitu CTCPLPT::Read_SR() {
	Bit8u const status = 0x9f | (ack ? 0x40 : 0);
	ack = false;
	return status;
}

void CTCPLPT::Write_PR (Bitu val) { datareg = (Bit8u)val; }
void CTCPLPT::Write_CON(Bitu val) {
	// init printer if bit 4 is switched on ...
	autofeed = !!(val & 0x02); // autofeed adds 0xa if 0xd is sent
    
	// data is strobed to the parallel printer on the falling edge of strobe bit
	if ((!(val        & 0x1)) &&
        ( (controlreg & 0x1))) {
		Putchar(datareg);
		if (autofeed && (datareg == 0xd))
            Putchar(0xa);
        
		ack = true;
	}
    
	controlreg = val;
}
void CTCPLPT::Write_IOSEL(Bitu /*val*/) { /* not needed for file printing functionality */  }

void CTCPLPT::handleUpperEvent(Bit16u /*type*/)
{
    if (!_socket) return;
    
    auto const timeoutExpire = lastUsedTick + timeout;
	if (timeoutExpire >= PIC_Ticks) {
		// Port has been touched in the meantime, try again later
		setEvent(0, float(timeoutExpire - PIC_Ticks) + 1);
        return;
	}
    
	handleClose();
}

void CTCPLPT::handleClose()
{
    lastChar = 0;
    _socket.reset();
}

bool CTCPLPT::OpenSocket()
{
    if (_socket) return true;
    
    _socket.reset(new TCPSocket(ip.c_str(), port));
    if (_socket->IsValid()) {
        setEvent(0, timeout);
        return true;
    }
    
    LOG_MSG("parallel%d: %s", port_nr + 1, _socket->ErrorGet().c_str());
    _socket.reset();
    
    return false;
}

